/*
   logtailer: a tool to watch log files.
   Copyright 2001, 2002, 2008, 2014 Adam Sampson <ats@offog.org>

   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
   copyright notice and this permission notice appear in all copies.

   THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
   WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
   ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
   ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/* Interval to poll the logs, in seconds. */
#define TIMEOUT 1

#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <curses.h>
#include <term.h>

void die(const char *format, ...) {
  va_list ap;

  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);

  exit(20);
}

typedef struct log_t {
  const char *name;
  const char *label;
  ino_t inode;
  off_t size;
  int fd;
  struct log_t *next;
} log_t;

log_t *log_list = NULL;
log_t *last_shown = NULL;

/* Print all the text that's available from a log. */
void printlog(log_t *log) {
  /* If the log is not the last one we printed from,
     then print a label. */
  if (log != last_shown) {
    vidattr(A_STANDOUT);
    fprintf(stdout, "<%s>\n", log->label);
    vidattr(A_NORMAL);
    last_shown = log;
  }

  /* Write everything we can to stdout. */
  while (true) {
    char buf[1024];
    int len = read(log->fd, buf, sizeof buf);
    if (len <= 0) break;

    fwrite(buf, 1, len, stdout);
  }
}

/* Open up a log file. */
bool openlog(log_t *log) {
  log->fd = open(log->name, O_RDONLY);
  if (log->fd < 0) return false;

  struct stat st;
  if (fstat(log->fd, &st) < 0) die("Unable to stat '%s'\n", log->name);
  if (!S_ISREG(st.st_mode)) die("'%s' is not a file\n", log->name);

  log->inode = st.st_ino;
  log->size = st.st_size;

  return true;
}

/* Add a logfile to the list. */
void addlog(const char *label, const char *name) {
  log_t *log = malloc(sizeof *log);
  log->label = label;
  log->name = name;

  if (openlog(log)) {
    /* Start at the end of the log. */
    lseek(log->fd, 0, SEEK_END);
  }

  log->next = log_list;
  log_list = log;
}

int main(int argc, char **argv) {
  int i = 1;
  while (i < argc) {
    if (!strcmp(argv[i++], "{")) {
      int labelpos = i++;
      while (strcmp(argv[i], "}")) {
        if (i + 1 >= argc) die("Invalid arguments\n");
        addlog(argv[labelpos], argv[i++]);
      }
      i++;
    } else {
      if (i >= argc) die("Invalid arguments\n");
      addlog(argv[i - 1], argv[i]);
      i++;
    }
  }
  if (i < argc) die("Too many arguments\n");
  if (!log_list) die("No logs specified\n");

  if (setupterm(NULL, STDOUT_FILENO, NULL) != OK) {
    die("Unable to initialise curses\n");
  }

  while (true) {
    bool found = false;

    /* Scan though all the fds. */
    for (log_t *log = log_list; log; log = log->next) {
      /* Try to open the log if it's currently closed. */
      if (log->fd == -1) {
        if (openlog(log)) {
          /* Read any data that's there already. */
          printlog(log);
          found = true;
        } else {
          /* Still not there -- ignore it for now. */
          continue;
        }
      }

      bool rotated = false;
      struct stat st;
      if (stat(log->name, &st) < 0) {
        /* The file doesn't exist any more. */
        rotated = true;
      } else if (st.st_ino != log->inode || st.st_size < log->size) {
        /* The file has been replaced. */
        rotated = true;
      } else if (st.st_size > log->size) {
        /* New data has been added; read it. */
        printlog(log);
        log->size = st.st_size;
        found = true;
      }

      if (rotated) {
        /* Read any remaining data, then close it. */
        printlog(log);
        close(log->fd);
        log->fd = -1;
        found = true;
      }
    }

    /* If we didn't find anything, wait for a while. */
    if (!found) sleep(TIMEOUT);
  }
}
